import time
from typing import Union, Optional
from base64 import b64encode
from random import random
from nacl.bindings.crypto_sign import crypto_sign

from umipy import bech32, Prefix


def to_public_key(address: str) -> list[int]:
    prefix, list_int, encoding = bech32.bech32_decode(address)
    public_key = bech32.convertbits(list_int, 5, 8, False)
    return public_key


def set_version(trx: list[int], version: int) -> None:
    trx.append(version)


def prefix_to_version(prefix: str) -> Optional[int]:
    if prefix == "genesis":
        return 0
    if len(prefix) not in (3, 5):
        return None

    to_check = []
    for i in range(len(prefix)):
        modifier = 47 if any(char.isdigit() for char in prefix[i]) else 96
        to_check.append(ord(prefix[i]) - modifier)

    if len(to_check) > 3:
        return 32768 + ((to_check[3] << 5) + to_check[4])

    return (to_check[0] << 10) + (to_check[1] << 5) + to_check[2]


def to_2(value: int) -> list[int]:
    return [(value >> 8) & 0xFF, value & 0xFF]


def to_4(value: int) -> list[int]:
    res = [
        (value & 0xFF000000) >> 24,
        (value & 0x00FF0000) >> 16,
        (value & 0x0000FF00) >> 8,
        (value & 0x000000FF) >> 0,
    ]
    return res


def set_list(trx: list[int], address: list[int]) -> None:
    trx += address


def set_amount(trx: list[int], amount: Union[int, float]) -> None:
    amount_to_bytes: bytes = int(amount).to_bytes(8, "big")
    to_list: list[int] = [int(i) for i in amount_to_bytes]
    trx += to_list


def sign(trx: list[int], sk: list[int]) -> None:
    signature = crypto_sign(bytes(trx), bytes(sk))
    trx += signature[:64]


def sign_transaction(trx: list[int], sk: list[int]) -> None:
    seconds = int(time.time()) - 10
    trx += to_4(seconds)
    trx += to_4(int(random() * 9999))
    trx.append(0)
    sign(trx, sk)


def transfer_coins(
    public_key: list[int],
    private_key: list[int],
    target_address: str,
    amount: Union[int, float],
    prefix: str | Prefix,
) -> str:
    trx: list[int] = []
    set_version(trx, 8)

    prefix_version = prefix_to_version(prefix)
    if prefix_version is None:
        return ""

    prefix_binary = to_2(prefix_version)
    from_addr = prefix_binary + public_key
    set_list(trx, from_addr)

    to_pk = to_public_key(target_address)
    to_addr = prefix_binary + to_pk
    set_list(trx, to_addr)

    set_amount(trx, amount * 100)

    sign_transaction(trx, private_key)

    return b64encode(bytes(trx)).decode()


def transfer_addresses(
    private_key: list[int],
    from_address: str,
    to_address: str,
    amount: Union[int, float],
) -> str:
    trx: list[int] = []
    set_version(trx, 8)

    token_to = to_address[:3]
    token_from = from_address[:3]

    prefix_binary_to = to_2(prefix_to_version(token_to))
    prefix_binary_from = to_2(prefix_to_version(token_from))

    from_pk = to_public_key(from_address)
    from_addr = prefix_binary_from + from_pk
    set_list(trx, from_addr)

    to_pk = to_public_key(to_address)
    to_addr = prefix_binary_to + to_pk
    set_list(trx, to_addr)

    set_amount(trx, amount * 100)

    sign_transaction(trx, private_key)

    return b64encode(bytes(trx)).decode()



BASE_URL_MAINNET = "https://mainnet.umi.top"
BASE_URL_TESTNET = "https://testnet.umi.top"


class Prefix:
    UMI = "umi"
    GLIZE = "glz"
    GLIZE_STAKING = "gls"

    ROD_STAKING = "rod01"
    ROD_TRADING = "rod"



import pydantic
from pydantic import AliasChoices


class BalanceResponse(pydantic.BaseModel):
    balance: float


class Transaction(pydantic.BaseModel):
    height: int
    block_timestamp: str = pydantic.Field(alias="blockTimestamp")
    block_height: int = pydantic.Field(alias="blockHeight")
    block_transaction_index: int = pydantic.Field(alias="blockTransactionIndex")
    hash: str
    type: str
    version: int
    amount: int
    sender_address: str = pydantic.Field(alias="senderAddress")
    sender_account_type: str = pydantic.Field(alias="senderAccountType")
    sender_account_balance: int = pydantic.Field(alias="senderAccountBalance")
    sender_account_interest_rate: int = pydantic.Field(
        alias="senderAccountInterestRate"
    )
    sender_account_transaction_count: int = pydantic.Field(
        alias="senderAccountTransactionCount"
    )
    recipient_amount: int = pydantic.Field(alias="recipientAmount")
    recipient_address: str = pydantic.Field(alias="recipientAddress")
    recipient_account_type: str = pydantic.Field(alias="recipientAccountType")
    recipient_account_balance: int = pydantic.Field(alias="recipientAccountBalance")
    recipient_account_interest_rate: int = pydantic.Field(
        alias="recipientAccountInterestRate"
    )
    recipient_account_transaction_count: int = pydantic.Field(
        alias="recipientAccountTransactionCount"
    )
    timestamp: str


class TransactionsResponse(pydantic.BaseModel):
    total_count: int = pydantic.Field(
        alias="totalCount", validation_alias=AliasChoices("totalCount", "total_count")
    )
    items: list[Transaction]


class Keys(pydantic.BaseModel):
    public_key: list[int]
    private_key: list[int]
    key_type: str


class WalletResponse(pydantic.BaseModel):
    mnemonic: str
    keys: Keys
    address: str



import base64
from typing import Optional, Union

import aiohttp
from nacl.bindings import crypto_sign, crypto_sign_open
from nacl.exceptions import BadSignatureError

from umipy.constants import Prefix, BASE_URL_MAINNET
from umipy.generate_wallet import generate_wallet, restore_wallet
from umipy.models import BalanceResponse, TransactionsResponse, WalletResponse, Keys
from umipy.transfer import transfer_coins, transfer_addresses, to_public_key


class UmiPy:
    def __init__(
        self, session: Optional[aiohttp.ClientSession] = None, base_url: str = BASE_URL_MAINNET
    ):
        self.base_url = base_url
        self.session = session or aiohttp.ClientSession()

    async def request(
        self,
        method: str,
        path: str,
        params: Optional[dict] = None,
        data: Optional[dict | str] = None,
    ):
        response = await self.session.request(
            method=method, url=f"{self.base_url}{path}", params=params, json=data
        )
        json_response = await response.json()
        return json_response

    async def close(self):
        await self.session.close()

    async def get_balance(self, address: str) -> BalanceResponse:
        response = await self.request("GET", f"/api/addresses/{address}/account")
        if "error" in response:
            return BalanceResponse(balance=0)

        return BalanceResponse(balance=response["data"]["confirmedBalance"] / 100)

    async def get_transactions(
        self, address: str, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> TransactionsResponse:
        params = {}
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset

        response = await self.request(
            "GET",
            f"/api/addresses/{address}/transactions",
            params=params,
        )

        if "error" in response:
            return TransactionsResponse(total_count=0, items=[])

        return TransactionsResponse(**response["data"])

    def generate_wallet(self, prefix: Prefix = Prefix.UMI) -> WalletResponse:
        address, mnemonic, public_key, private_key = generate_wallet(prefix=prefix)
        return WalletResponse(
            address=address,
            mnemonic=mnemonic,
            keys=Keys(
                public_key=public_key, private_key=private_key, key_type="банан беби"
            ),
        )

    async def send(
        self,
        private_key: list[int],
        public_key: list[int],
        target_address: str,
        amount: Union[float, int],
        prefix: Prefix = Prefix.UMI,
    ) -> bool:
        encoded_data = transfer_coins(
            public_key=public_key,
            private_key=private_key,
            target_address=target_address,
            amount=amount,
            prefix=prefix,
        )

        response = await self.request(
            method="POST", path=f"/api/mempool", data={"data": encoded_data}
        )
        if "error" in response:
            return False

        return response["data"]

    async def send_addresses(
        self,
        private_key: list[int],
        from_address: str,
        target_address: str,
        amount: Union[float, int],
    ) -> bool:
        encoded_data = transfer_addresses(
            private_key=private_key,
            from_address=from_address,
            to_address=target_address,
            amount=amount,
        )
        response = await self.request(
            method="POST", path=f"/api/mempool", data={"data": encoded_data}
        )
        if "error" in response:
            return False

        return response["data"]

    def restore_wallet(
        self,
        mnemonic: str,
        prefix: Prefix = Prefix.UMI,
    ) -> WalletResponse:
        address, mnemonic, public_key, private_key = restore_wallet(
            mnemonic=mnemonic, prefix=prefix
        )
        return WalletResponse(
            address=address,
            mnemonic=mnemonic,
            keys=Keys(
                public_key=public_key, private_key=private_key, key_type="банан беби"
            ),
        )

    def sign_message(
        self,
        private_key: list[int],
        message: str,
    ) -> str:
        encoded_message = message.encode()
        result = crypto_sign(message=encoded_message, sk=bytes(private_key))

        return base64.b64encode(result[: -len(encoded_message)]).decode()

    def sign_verify(self, signature: str, original_message: str, address: str) -> bool:
        public_key = to_public_key(address)
        base64_sig = base64.b64decode(signature)

        base64_sig += original_message.encode()
        try:
            result = crypto_sign_open(signed=base64_sig, pk=bytes(public_key))
        except BadSignatureError:
            return False

        return result.decode() == original_message



from typing import Union
from hashlib import pbkdf2_hmac, sha256
from mnemonic import Mnemonic

from nacl.bindings import crypto_sign_seed_keypair

from umipy import bech32, Prefix


def generate_mnemo_words() -> str:
    mnemo = Mnemonic("english")
    return mnemo.generate(strength=256)


def to_seed(words: str) -> bytes:
    return pbkdf2_hmac("sha512", words.encode(), "mnemonic".encode(), 2048)


def to_unsigned_list_int(seed: bytes) -> list[int]:
    to_sha256 = sha256(seed).hexdigest()
    return [int(to_sha256[i : i + 2], 16) for i in range(0, len(to_sha256), 2)]


def generate_pk_sk(unsigned_list_int: list[int]) -> tuple[list[int], list[int]]:
    public_key, secret_key = crypto_sign_seed_keypair(bytes(unsigned_list_int))
    return [int(i) for i in public_key], [int(i) for i in secret_key]


def generate_wallet_address(public_key: Union[list[int], bytes], prefix: str) -> str:
    convert_bits = bech32.convertbits(public_key, 8, 5, True)
    return bech32.bech32_encode(prefix, convert_bits, 0)


def generate_wallet(prefix: str | Prefix) -> tuple[str, str, list[int], list[int]]:
    mnemo = generate_mnemo_words()
    seed = to_seed(mnemo)
    unsigned_int = to_unsigned_list_int(seed)
    pk, sk = generate_pk_sk(unsigned_int)
    address = generate_wallet_address(pk, prefix=prefix)
    return address, mnemo, pk, sk


def restore_wallet(mnemonic: str, prefix: str | Prefix):
    seed = to_seed(mnemonic)
    unsigned_int = to_unsigned_list_int(seed)
    pk, sk = generate_pk_sk(unsigned_int)
    address = generate_wallet_address(pk, prefix=prefix)
    return address, mnemonic, pk, sk
